#include "control.h"

#include "objects.h"
//#include "math3d.h"

#include <cstdlib>
#include <cassert>

using namespace std;

#define distance(A, B)       (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y])+((A)[Z]-(B)[Z])*((A)[Z]-(B)[Z]) ))

//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//


//--------------------------------


cController::cController()
{
  mDevice = NULL;
  mThrottle = 0;
  mWeapon = 0;
}


void cController::processSelf()
{
  transferSelf();
}


void cController::transferSelf()
{
  if (mDevice == NULL) return;
/*
  mDevice->ChasisLR(mChasis.mV[Y]);  // around [Y]-axis
  mDevice->ChasisUD(mChasis.mV[X]);
  mDevice->TowerLR(mTower.mV[Y]);
  mDevice->TowerUD(mTower.mV[X]);
  mDevice->Throttle(mThrottle);
*/
  mDevice->handleMsg(0, (long) "ChasisLR", (long) mChasis.mV[Y], 0);
  mDevice->handleMsg(0, (long) "ChasisUD", (long) mChasis.mV[X], 0);
  mDevice->handleMsg(0, (long) "TowerLR", (long) mTower.mV[Y], 0);
  mDevice->handleMsg(0, (long) "TowerUD", (long) mTower.mV[X], 0);
  mDevice->handleMsg(0, (long) "Throttle", (long) mThrottle, 0);
}


//--------------------------------


cAI::cAI()
{
  mDisturb = 0.5;
  clearState();
}


void cAI::clearState()
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  mNav = 1;
//  overideState(WAIT, (char*) NULL);
}


void cAI::overideState(int state, cEntity *entity)
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;
  if (entity == NULL) return;
  if (entity->getFragged()) return;

  state_struct* st;

	if (!mState.empty())
	{
    st = mState.top();
    mState.pop();
    delete st;
	};

  st = new state_struct;
  st->state = state;
  st->entity = entity;
  st->entity_name = entity->getName();
  mState.push(st);
}


void cAI::overideState(int state, char* entityname)
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;

  if (entityname == NULL) return;

  state_struct* st;

	if (!mState.empty())
	{
    st = mState.top();
    mState.pop();
    delete st;
	};

  st = new state_struct;
  st->state = state;
  st->entity = NULL;
  st->entity_name = entityname;
  mState.push(st);
}


void cAI::newState(int state, cEntity *entity)
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;

  if (entity == NULL) return;
  else if (entity->getFragged()) return;

  state_struct* st;

  st = new state_struct;
  st->state = state;
  st->entity = entity;
  st->entity_name = entity->getName();
  mState.push(st);
}


void cAI::newState(int state, char* entityname)
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;

  if (entityname == NULL) return;

  state_struct* st;

  st = new state_struct;
  st->state = state;
  st->entity = NULL;
  st->entity_name = entityname;
  mState.push(st);
}


void cAI::oldState()
{
//  if (mDevice != NULL) mThreshold = mDevice->mDamage + mDisturb;

	// Is there any state to remove?
  if (mState.size() == 0) return;

  state_struct* st;

	// Remove state.
  st = mState.top();
  mState.pop();
  delete st;

	if (mState.size() > 0)
	{
    st = mState.top();
    st->entity = NULL; // Pointer might be invalid, search again.
	};
}


void cAI::processSelf()
{
  if (mDevice == NULL) return;

  if (mState.empty())
	{
		waitEnemy();
	  // Transfer control wishes to the controlled device.
    transferSelf();
		return;
	};

//  if (mState.empty()) return;
  state_struct* st = mState.top();

  if ((st->entity == NULL) && (st->entity_name.size() > 1))
  {
	  // Search world for the entity name.
	  cObjectFunction* objfun = new cObjectFunction_findName(NULL, (char*) st->entity_name.c_str(), 1);
//	  cObjectFunction* objfun = new cObjectFunction_findName(NULL, "Bravo", 1); // Bug: Can only find self?!?
    mDevice->getWorld()->callObjects(objfun, 0, true);

	  // if not available then pop state and return.
    if (objfun->collection.empty()) {
		  objfun->collection.clear();
	    delete objfun;
      oldState();
      return;
    };
    st->entity = (cEntity*) objfun->collection.front();
		delete objfun;
  };

	// Set Self Origin Vector.
  mSelf = mDevice->mPos;

	// Set Target Destination Vector.
  if (st->entity != NULL) {
    mDest = st->entity->mPos;
  };

  switch(st->state) {
    case WAIT: waitEnemy();      break;
    case ENGAGE: engageEnemy();    break;
    case TARGET: targetEnemy();    break;
    case ATTACK:  attackEnemy();    break;
    case FOLLOW: followLeader();   break;
    case FLEE: fleeEnemy();      break;
    case ROUTE: routeNav();       break;
    case PATROL: patrolArea();     break;
    default: report("Invalid State Request!\n");     break;
  };

	// Transfer control wishes to the controlled device.
  transferSelf();

	// Reset control wishes.
  mTower.Set(0, 0, 0);
  mChasis.Set(0, 0, 0);
}


void cAI::waitEnemy()
{
  // Search world for any enemy in range.
	cObjectFunction* objfun = new cObjectFunction_rangeScan(mDevice->mPos.mV, 50, 1);
	objfun->caller = mDevice;
  mDevice->getWorld()->callObjects(objfun, 1, false);

  if (!objfun->collection.empty()) {
		cEntity* ent = (cEntity*) objfun->collection.front();
		printf("%s Targets %s\n", mDevice->getName().c_str(), ent->getName().c_str());
		assert(objfun->collection.front() != NULL);
		assert(objfun->collection.front() != mDevice);
    newState(TARGET, ent);
		objfun->collection.clear();
	};
  delete objfun;

  mThrottle = 0.0f;
}


void cAI::engageEnemy()
{
  float d = lineDistance(mSelf.mV, mDest.mV);

	// Target close.
  if (d < 55.0)
  {
		printf("%s Targets %s\n", mDevice->getName().c_str(), mState.top()->entity->getName().c_str());
    newState(TARGET, mState.top()->entity);
    return;
  };

  cVector dir = mSelf;    dir.Sub(mDest);
//  cVector pol = dir;      pol.Polar();
  float  rd = 0.0f;
  int i;

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mThrottle = 100.0f * (1.0f - fabs(rd) / 360.0f);

  mChasis.mV[Y] = rd;

  float *tower = NULL;
  mDevice->handleMsg(0, (long) "getTowerPtr", (long) &tower, 0);
  assert(tower != NULL);

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + tower[Y]);
//  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + mDevice->mTwr.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mTower.mV[Y] = rd;

//  pol.mV[X] -= 180;
//  pol.mV[X] = fabs(pol.mV[X]);
//  float  rd2 = (pol.mV[X] - mDevice->mTwr.mV[X]-90);
  
//  mTower.mV[X] = rd2;
}


void cAI::targetEnemy()
{

  float d = lineDistance(mSelf.mV, mDest.mV);

	// Target close.
  if (d < 16.0)
  {
		printf("%s Attacks %s\n", mDevice->getName().c_str(), mState.top()->entity->getName().c_str());
    newState(ATTACK, mState.top()->entity);
    return;
  };

	// Target Fleed/ out of range.
  if (d > 60.0)
  {
		printf("%s  can't Target %s\n", mDevice->getName().c_str(), mState.top()->entity->getName().c_str());
    oldState(); // Back to whatever was before (possibly engage or wait).
    return;
  };

  cVector dir = mSelf;    dir.Sub(mDest);
//  cVector pol = dir;      pol.Polar();
  float  rd = 0.0f;
  int i;

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mThrottle = 100.0f * (1.0f - fabs(rd) / 360.0f);

  mChasis.mV[Y] = rd;

  float *tower = NULL;
  mDevice->handleMsg(0, (long) "getTowerPtr", (long) &tower, 0);
  assert(tower != NULL);

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + tower[Y]);
//  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + mDevice->mTwr.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mTower.mV[Y] = rd;

//  pol.mV[X] -= 180;
//  pol.mV[X] = fabs(pol.mV[X]);
//  float  rd2 = (pol.mV[X] - mDevice->mTwr.mV[X]-90);
  
//  mTower.mV[X] = rd2;

  if (rand()%6 == 0) {
    if (fabs(rd) < 3) {
      mDevice->handleMsg(0, (long) "fireall", (long) (mState.top()->entity), 0);
    };
  };
}


void cAI::attackEnemy()
{
//  static float  tf = 0;
//  tf += 3;

  float  d = lineDistance(mSelf.mV, mDest.mV);

  if (d > 20.0)
  {
		printf("%s can't Attack %s\n", mDevice->getName().c_str(), mState.top()->entity->getName().c_str());
    oldState(); // back to target
    return;
  };


  mThrottle = 50;

  cVector dir = mSelf;    dir.Sub(mDest);
//  cVector pol = dir;      pol.Polar();
  float  rd = 0.0f;
  int i;

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mChasis.mV[Y] = rd;

  mThrottle = 100.0f * (1.0f - fabs(rd) / 360.0f);
  float f = (d - 11);
  if (f < -5) f = -5;
  if (f > +10) f = +10;
  mThrottle *= f * 0.1;

  float *tower = NULL;
  mDevice->handleMsg(0, (long) "getTowerPtr", (long) &tower, 0);
  assert(tower != NULL);
  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + tower[Y]);
//  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + mDevice->mTwr.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mTower.mV[Y] = rd;


//  pol.mV[X] -= 180;
//  pol.mV[X] = fabs(pol.mV[X]);
//  float  rd2 = (pol.mV[X] - mDevice->mTwr.mV[X]-90);

//  mTower.mV[X] = rd2;

  if (rand()%6 == 0) {
    if (fabs(rd) < 3) {
      mDevice->handleMsg(0, (long) "fireall", (long) (mState.top()->entity), 0);
    };
  };

//  ((cMech*)mDevice)->Pattern(tf, "nrnlln");
}


void cAI::followLeader()
{
}


void cAI::fleeEnemy()
{
}


void cAI::routeNav()
{
/*
  mDest = mObjectStack[mStatePointer]->mParts[mNav]->mPos;

  float  d = lineDistance(mSelf.mV, mDest.mV);

  if (d <= 1.0) {
    mNav++;
    if (mNav >= (mObjectStack[mStatePointer]->mNParts)) {
      mNav = 1;
      oldState();
    };
  };

  cVector dir = mSelf;    dir.Sub(mDest);
//  cVector pol = dir;      pol.Polar();
  float  rd;

  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mChasis.mV[Y] = rd;

  mThrottle = 100.0f * (1.0f - fabs(rd) / 360.0f);

  float *tower = NULL;
  mDevice->handleMsg(0, (long) "getTowerPtr", (long) &tower, 0);
  assert(tower != NULL);
  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + tower[Y]);
//  rd = (180 + atan2(dir.mV[X], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[Y] + mDevice->mTwr.mV[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.

  mTower.mV[Y] = rd;

//  float rd2 = (180 + atan2(dir.mV[Y], dir.mV[Z]) * 180.0f/3.145479f) - (mDevice->mRot.mV[X] + mDevice->mTwr.mV[X]);
//  if (fabs(rd2) > 180) rd2 = -(rd2 / fabs(rd2)) * (360 - fabs(rd2)); // -sign(rd) * circle other way around.

//  mTower.mV[X] = rd2;

//  fprintf(stderr, "Damage: %f\n", mDevice->mDamage);
*/
}


void cAI::patrolArea()
{
}


