#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "math3d.h"

#include <time.h>
#include <stdio.h>

#include "envir.h"

#include <cmath>

#ifndef M_PI
#define M_PI 3.14159265359
#endif

#undef xmod
#undef xabs

void cDome::DrawStuff(cMatrix camera, SCALAR x, SCALAR z, SCALAR s, DWORD tx)
{
  glColor4f(1.0, 1.0, 1.0, 0.6);
  cMatrix cm;
  cm.MovRel(x, 0.0, z);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0);
    glVertex3f(+s, 0.0, -s);
    glTexCoord2f(1.0, 0.0);
    glVertex3f(-s, 0.0, -s);
    glTexCoord2f(1.0, 1.0);
    glVertex3f(-s, 0.0, +s);
    glTexCoord2f(0.0, 1.0);
    glVertex3f(+s, 0.0, +s);
  glEnd();
  glEnable(GL_DEPTH_TEST);
}

/*
void cDome::DrawClouds(cMatrix camera, DWORD tx)
{
  float m[16];
  float p[3];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  // Copy Position from 4th Row.
  p[0] = m[3*4+0];
  p[1] = m[3*4+1];
  p[2] = m[3*4+2];

  glPushMatrix();
  glTranslatef(p[0], 0, p[2]);

  clock_t c = clock();
  float e = 20.0f * M_PI / 180.0f;
  float h = 200.0f;
  float t = 0.01f * (float(c) / CLOCKS_PER_SEC);
  t = fmod(t, 1);

  glColor4fv(mLoColor.mV);
  glFogfv(GL_FOG_COLOR, mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  float dx = 1.0f * t;
  float dz = 1.0f * t;
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, h, 0.0);
  for (int i=0; i<=18; i++) {
    glTexCoord2f(cos(i*e)*5.0+dx, sin(i*e)*5.0+dz);
    glVertex3f(cos(i*e)*2000.0, h, sin(i*e)*2000.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);

  glPopMatrix();
}
*/
void cDome::DrawClouds(cMatrix camera, DWORD tx)
{
  clock_t c = clock();
  SCALAR e = 20.0*M_PI/180.0f;
  SCALAR h = 200.0;
  SCALAR t = 0.01 * (SCALAR(c) / CLOCKS_PER_SEC);
  t = fmod(t, 1);
  INTEG i;
  SCALAR dx, dz;
  cVector p;
  cMatrix cm;
  cMatrix n = camera;
  n.Inv();
  n.Apply(p);
  cm.MovRel(0.0, p.mV[Y]*1.0, 0.0);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  glColor4fv(mLoColor.mV);
  glFogfv(GL_FOG_COLOR, mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  dx = 1.0*t;
  dz = 1.0*t;
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, h, 0.0);
  for (i=0; i<=18; i++) {
    glTexCoord2f(cos(i*e)*5.0+dx, sin(i*e)*5.0+dz);
    glVertex3f(cos(i*e)*2000.0, h, sin(i*e)*2000.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);
}


/*
void cDome::DrawGround(cMatrix camera, DWORD tx)
{
  float m[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  // Copy Position from 4th Row.
  float p[3];
  p[0] = m[3*4+0];
  p[1] = m[3*4+1];
  p[2] = m[3*4+2];

//  p[1] = 

  // 4th Row = 0.
  m[0*4+3] =
  m[1*4+3] =
  m[2*4+3] = 0.0f;

  // 4th Col = 0.
  m[3*4+0] = m[3*4+1] = m[3*4+2] = 0.0f;
 
  m[3*4+3] = 1.0f;

  glPushMatrix();
  glLoadIdentity();
  glTranslatef(0, -5, 0);
//  glTranslatef(0, p[Y], 0);
  glMultMatrixf(m);

//  float dx = 0, dz = 0;
  float dx = p[X]/5.0, dz = p[Z]/5.0;
  glColor4fv(mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0, 1, 0);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, 0.0, 0.0);
  float e = 20.0*M_PI/180.0f;
  for (int i=0; i<=18; i++) {
    glNormal3f(0, 1, 0);
    glTexCoord2f(cos(-i*e)*500.0+dx, sin(-i*e)*500.0+dz);
    glVertex3f(cos(-i*e)*2500.0, 0, sin(-i*e)*2500.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);

  glPopMatrix();
}
*/
void cDome::DrawGround(cMatrix camera, DWORD tx)
{
  INTEG i;
  cVector p;
  cMatrix cm;
  cMatrix n = camera;
  n.Inv();
  n.Apply(p);
  cm.MovRel(p.mV[X]*1.0, 0.0, p.mV[Z]*1.0);
  cm.Mul(camera);
  SCALAR dx = p.mV[X]/5.0, dz = p.mV[Z]/5.0;
  glLoadMatrixf(cm.mM);
  glColor4fv(mLoColor.mV);
  glDisable(GL_DEPTH_TEST);
  glBindTexture(GL_TEXTURE_2D, tx);
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0, 1, 0);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, 0.0, 0.0);
  SCALAR e = 20.0*M_PI/180.0f;
  for (i=0; i<=18; i++) {
    glNormal3f(0, 1, 0);
    glTexCoord2f(cos(-i*e)*500.0+dx, sin(-i*e)*500.0+dz);
    glVertex3f(cos(-i*e)*2500.0, 0.0, sin(-i*e)*2500.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);
}


void cDome::DrawDome(cMatrix camera)
{
  float m[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  // 4th Row = 0.
  m[0*4+3] =
  m[1*4+3] =
  m[2*4+3] = 0.0f;

  // 4th Col = 0.
  m[3*4+0] = m[3*4+1] = m[3*4+2] = 0.0f;
 
  m[3*4+3] = 1.0f;

  glPushMatrix();
  glLoadIdentity();
  glMultMatrixf(m);

  int steps = 8;
  float y1, y2, s1, s2, f = 90.0/steps*0.017454;
  cVector color1 = mUpColor, color2 = mUpColor;
  cVector slope = mLoColor;
  slope.Sub(mUpColor);
  slope.Div((SCALAR)steps);

  glDisable(GL_DEPTH_TEST);
  s1 = sin(0*f);
  y1 = cos(0*f)*100;
  for (int i=1; i<=steps; i++) {
    s2 = sin(i*f);
    y2 = cos(i*f)*100;
    color2.Add(slope);
    glBegin(GL_TRIANGLE_STRIP);
    for (int j=0; j<=18; j++) {
      glColor3fv(color1.mV);
      glNormal3f(s1*cos(j*20*0.017454), y1*0.01f, s1*sin(j*20*0.017454));
      glVertex3f(s1*cos(j*20*0.017454)*100, y1, s1*sin(j*20*0.017454)*100);
      glColor3fv(color2.mV);
      glNormal3f(s2*cos(j*20*0.017454), y2*0.01f, s2*sin(j*20*0.017454));
      glVertex3f(s2*cos(j*20*0.017454)*100, y2, s2*sin(j*20*0.017454)*100);
    };
    glEnd();
    color1 = color2;
    y1 = y2;
    s1 = s2;
  };
  glEnable(GL_DEPTH_TEST);

  glPopMatrix();
}
/*
void cDome::DrawDome(cMatrix camera)
{
  cVector p;
  cMatrix n = camera;
  cMatrix cm;
  n.Inv();
  n.Apply(p);
  cm.MovRel(p.mV[X]*1.0, p.mV[Y]*1.0, p.mV[Z]*1.0);
  cm.Mul(camera);
  glLoadMatrixf(cm.mM);
  INTEG i, j, steps = 8;
  SCALAR y1, y2, s1, s2, f = 90.0/steps*0.017454;
  cVector color1 = mUpColor, color2 = mUpColor;
  cVector slope = mLoColor;
  slope.Sub(mUpColor);
  slope.Div((SCALAR)steps);
  glDisable(GL_DEPTH_TEST);
  s1 = sin(0*f);
  y1 = cos(0*f)*100;
  for (i=1; i<=steps; i++) {
    s2 = sin(i*f);
    y2 = cos(i*f)*100;
    color2.Add(slope);
    glBegin(GL_TRIANGLE_STRIP);
    for (j=0; j<=18; j++) {
      glColor3fv(color1.mV);
      glVertex3f(s1*cos(j*20*0.017454)*100, y1, s1*sin(j*20*0.017454)*100);
      glColor3fv(color2.mV);
      glVertex3f(s2*cos(j*20*0.017454)*100, y2, s2*sin(j*20*0.017454)*100);
    };
    glEnd();
    color1 = color2;
    y1 = y2;
    s1 = s2;
  };
  glEnable(GL_DEPTH_TEST);
}
*/

void cDome::SetDayTime(SCALAR hour)
{
  SCALAR t = ((hour * 360.0 / 24.0) - 90.0);
  SCALAR s = sin(t*0.017454)*0.5+0.5;
  mUpColor.Set(0.0, 0.0, s*0.8+0.2, 1.0);
  mLoColor.Set(s*0.95+0.05, s*0.95+0.05, s*0.95+0.05, 1.0);
  mT = hour;
}


void cDome::SetSystemDayTime()
{
  time_t t = time(NULL);
  tm *l = localtime(&t);
  SCALAR lt = (l->tm_hour + l->tm_min / 60.0 + l->tm_sec / 3600.0);
  SetDayTime(lt);

//  SetDayTime((l->tm_sec)*24.0/60.0);
// ^- 60 seconds equals 24 hours - good for checking day-colors
}



