#define nUSE_3DFX
#define USE_GL

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#ifndef _WIN32
#ifdef USE_3DFX
#include <GL/xmesa.h>
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <fcntl.h>

#ifndef _WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#include <cassert>

class cInput {
  public:
  int buttons, x, y;
};

const unsigned int JS_RETURN = sizeof(cInput);

char        *js0name = "/dev/js0",   *js1name = "/dev/js1";
int                       js0file,                 js1file;
int                     js0status,               js1status;
int                         js0xm,                   js0ym;
int                         js1xm,                   js1ym;
cInput                      js0,                     js1;
cInput                      kb0,                     kb1;


//
// Misc.
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//


#include "math3d.h"
#include "util3d.h"
#include "pose3d.h"
#include "conio3d.h"
#include "viewer3d.h"

#ifndef _WIN32
#include <GL/glut.h>
#endif

#ifdef _WIN32
#include <GL\glut.h>
#endif

#include "objects.h"
#include "control.h"
#include "envir.h"

#define FIRE_KEY ' '
#define JUMPJET_KEY 'j'
#define CAMERA_KEY 'c'
#define SAT_CAMERA_KEY 'm'
#define WIREFRAME_KEY 'w'
#define PAUSE_KEY 'p'
#define FXMODE_KEY 'f'



void Fatal(char *errorString)
{
#ifdef USE_3DFX
  XMesaSetFXmode(XMESA_FX_WINDOW);
#endif
  fprintf(stderr, "%s\n", errorString);
  fflush(stderr);
  exit(1);
}


BOOLEAN LoadTgaTex(char *fname, BOOLEAN transparency, BOOLEAN mipmap, WORD &TexID)
{
  cBmp *b = new cBmp;
  cVector pink;

  if (b->LoadTga(fname)) return 1;
  if (transparency) {
    pink.Set(255, 0, 255);
    b->MakeAlpha(pink);
    b->Mul(W, 0.7f); // make more transparent
  };
  return b->glRegister(mipmap, TexID);
  delete [] b;
}




cConio3d             *con[2], *cDummy;

cViewer               mainView;

cWorld               *planet;

cController          *theUser;


cWeapon              *MGun;

cProjectile          *theShot;

//cNavSequence         *navs;


int                   external_camera = 0;
int                   satelite_camera = 0;

int                   width       = 0;
int                   height      = 0;

INTEG                 rButton     = 0;
INTEG                 mButton     = 0;
INTEG                 lButton     = 0;
SCALAR                xrSpeed     = 0;
SCALAR                yrSpeed     = 0;
SCALAR                mxpos       = 0;
SCALAR                mypos       = 0;
BOOLEAN               paused      = 0;
BOOLEAN               fullscreen  = 1;
BOOLEAN               js_disabled = 0;
BOOLEAN               wireframe   = 0;

WORD                  TexID[100];

char                  stringBuffer[512];


void Init()
{
  INTEG i, j;
  float Black[3] = { 0.0, 0.0, 0.1 };
  float lightColor[3] = { 0.5, 0.2, 1.0 };
  float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
  float fogDensity = 0.001;

//  glDisable(GL_DITHER);
  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  glClearColor(0.0, 0.125, 0.0, 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//  glShadeModel(GL_FLAT);
  glShadeModel(GL_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
  glDepthFunc(GL_LEQUAL);

  glFogi(GL_FOG_MODE, GL_EXP);
  glFogfv(GL_FOG_COLOR, fogColor);
  glFogf(GL_FOG_DENSITY, fogDensity);

//  glHint(GL_FOG_HINT,GL_FASTEST);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
//  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

  if (
    LoadTgaTex("data/textures/house0.tga",    0,1, TexID[0]) |
    LoadTgaTex("data/textures/clouds0.tga",   1,1, TexID[3]) |
    LoadTgaTex("data/textures/font0.tga",     1,0, TexID[4]) |
    LoadTgaTex("data/textures/flare0.tga",    1,1, TexID[5]) |
    LoadTgaTex("data/textures/tile0.tga",     0,1, TexID[6]) |
    LoadTgaTex("data/textures/ground0.tga",   1,1, TexID[7]) |
    LoadTgaTex("data/textures/tree0.tga",     1,1, TexID[8]) 
  ) Fatal("could not load tga texture");

  con[0] = new cConio3d;
  con[0]->Resize(16, 8, 0.02);
  con[0]->SetFont(TexID[4]);
  con[0]->Insert("--------------\n");  
  con[0]->Insert("Lin Warrior 3D\n");  
  con[0]->Insert("--------------\n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("3D Console 0  \n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("Status: INACTIVE\n");  
  con[0]->mPose.mPos.mV[X] = +0.29;
  con[0]->mPose.mPos.mV[Y] = +0.27;
  con[0]->mPose.mRot.mV[X] = +25;
  con[0]->mPose.mRot.mV[Y] = -25;
  con[0]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  con[1] = new cConio3d;
  con[1]->Resize(16, 8, 0.02);
  con[1]->SetFont(TexID[4]);
  con[1]->Insert("--------------\n");  
  con[1]->Insert("Lin Warrior 3D\n");  
  con[1]->Insert("--------------\n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("3D Console 1  \n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("Status: ACTIVE\n");  
  con[1]->mPose.mPos.mV[X] = -0.29;
  con[1]->mPose.mPos.mV[Y] = +0.27;
  con[1]->mPose.mRot.mV[X] = +25;
  con[1]->mPose.mRot.mV[Y] = +25;
  con[1]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  cDummy = con[0];



  FILE *file;
  planet = new cWorld;


  cModel *mod;
/*
  mod = new cDummyModel;
  mod->setName("cDummyModel");
  planet->spawnObject(mod);
*/
  mod = new cParticleModel;
  mod->setName("cParticleModel");
  planet->spawnObject(mod);

  mod = new cConstructModel;
  mod->setName("cConstructModel");
  planet->spawnObject(mod);

  mod = new cPlantModel;
  mod->setName("cPlantModel");
  planet->spawnObject(mod);


  float pos[3];
  float rot[3];

  cMech* mech;
  cAI* ai;

  MGun = new cWeapon;



#if 1
  pos[0] = -50;
  pos[1] = 0;
  pos[2] = -50;
  mech = new cMech(pos);
  mech->setName("Alpha");
  assert(mech != NULL);

  theUser = new cController;
  theUser->mDevice = mech;
  mech->setController(theUser);

  planet->spawnObject(mech);

#if 1
  mech->mountWeapon("LTorsor", MGun);
  mech->mountWeapon("RTorsor", MGun);
  mech->mountWeapon("RLoArm",  MGun);
  mech->mountWeapon("LLoArm",  MGun);
#endif
#endif



#if 1
  pos[0] = 50;
  pos[1] = 0;
  pos[2] = -50;
  mech = new cMech(pos);
  mech->setName("Bravo");
  assert(mech != NULL);

  ai = new cAI;
  ai->mDevice = mech;
  mech->setController(ai);
//  ai->newState(cAI::TARGET, "Charly");
//  ai->newState(cAI::ROUTE, "Attackroute");

  planet->spawnObject(mech);

#if 1
//  mech->mountWeapon("LLoArm",  MGun);
  mech->mountWeapon("RLoArm",  MGun);
#endif
#endif




#if 1
  pos[0] = 50;
  pos[1] = 0;
  pos[2] = 50;
  mech = new cMech(pos);
  mech->setName("Charly");
  assert(mech != NULL);

  ai = new cAI;
  ai->mDevice = mech;
  mech->setController(ai);
//  ai->newState(cAI::WAIT, (char*) NULL);
//  ai->newState(cAI::TARGET, "Delta");
//  ai->newState(cAI::ROUTE, "Attackroute");

  planet->spawnObject(mech);

#if 1
  mech->mountWeapon("LTorsor",  MGun);
//  mech->mountWeapon("RTorsor",  MGun);
#endif
#endif



#if 1
  pos[0] = -50;
  pos[1] = 0;
  pos[2] = 50;
  mech = new cMech(pos);
  mech->setName("Delta");
  assert(mech != NULL);

  ai = new cAI;
  ai->mDevice = mech;
  mech->setController(ai);
  ai->newState(cAI::ENGAGE, "Charly");
//  ai->newState(cAI::ROUTE, navs);

  planet->spawnObject(mech);

#if 1
//  mech->mountWeapon("LTorsor",  MGun);
  mech->mountWeapon("RTorsor",  MGun);
#endif
#endif



/*
  navs = new cNavSequence;
  file = fopen("data/objects/navs0.txt", "r");
  if (file==NULL) Fatal("Unable To Open Navs File...Abort.");
  if (navs->loadSelf(file)) Fatal("Invalid Navs File...Abort");
  fclose(file);
  navs->setName("City Route");
*/

#if 0
  int bld_dist = 50;
  cBuilding* bld;
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      if (!((i==1) && (j==1))) {
        pos[0] = i*bld_dist*2-bld_dist;
        pos[1] = 0;
        pos[2] = j*bld_dist*2-bld_dist;
        bld = new cBuilding(pos);
        planet->spawnObject(bld);
      };
    };
  };
#endif

#if 1
  int dist = 320;
  float a, b;
  cTree* tr;
/*
  for (i=0; i<20; i++) {
    a = rand()%360;
//    b = dist + rand() % (2*dist);
    b = rand() % (2*dist);
    pos[0] = sin(a*0.017454)*b;
    pos[1] = 0;
    pos[2] = cos(a*0.017454)*b;
    rot[0] = 10-rand()%20;
    rot[1] = 10*(rand()%36);
    rot[2] = 10-rand()%20;
    tr = new cTree(pos, rot);
    planet->spawnObject(tr);
  };
*/
  for (i=0; i<9; i++) {
    a = dist - rand() % (2*dist);
    b = dist - rand() % (2*dist);
    pos[0] = a;
    pos[1] = 0;
    pos[2] = b;
    rot[0] = 0;//10-rand()%20;
    rot[1] = 10*(rand()%36);
    rot[2] = 0;//10-rand()%20;
    tr = new cTree(pos, rot, "2 2.0");
    planet->spawnObject(tr);
  };

  for (i=0; i<6; i++) {
    a = dist - rand() % (2*dist);
    b = dist - rand() % (2*dist);
    pos[0] = a;
    pos[1] = 0;
    pos[2] = b;
    rot[0] = 0;//10-rand()%20;
    rot[1] = 10*(rand()%36);
    rot[2] = 0;//10-rand()%20;
    tr = new cTree(pos, rot, "3 2.0");
    planet->spawnObject(tr);
  };

  for (i=0; i<3; i++) {
    a = dist - rand() % (2*dist);
    b = dist - rand() % (2*dist);
    pos[0] = a;
    pos[1] = 0;
    pos[2] = b;
    rot[0] = 0;//10-rand()%20;
    rot[1] = 10*(rand()%36);
    rot[2] = 0;//10-rand()%20;
    tr = new cTree(pos, rot, "4 2.0");
    planet->spawnObject(tr);
  };

  for (i=0; i<0; i++) {
    a = dist - rand() % (2*dist);
    b = dist - rand() % (2*dist);
    pos[0] = a;
    pos[1] = 0;
    pos[2] = b;
    rot[0] = 0;//10-rand()%20;
    rot[1] = 10*(rand()%36);
    rot[2] = 0;//10-rand()%20;
    tr = new cTree(pos, rot, "5 2.0");
    planet->spawnObject(tr);
  };
#endif

/*
  mech->mTexture      = TexID[mech->mAttrib];
  navs->mTexture      = TexID[mech->mAttrib];
*/

  planet->mDome = new cDome;
//  planet->mFireTx = TexID[5];

//  planet->spawnObject(navs);


#define LIGHTON

#ifdef LIGHTON
  glEnable(GL_LIGHTING);
  float p[] = { 0, 100, 0, 1 };
  glLightfv(GL_LIGHT0, GL_POSITION, p);
  glEnable(GL_COLOR_MATERIAL);
#endif
}


void Reshape(int w, int h)
{
  width = w;
  height = h;

#ifdef USE_3DFX
  if (width > 640) width = 640;
  if (height > 480) height = 480;
#endif

  mainView.mType = 0;
  mainView.Move(0, 0);
  mainView.Resize(width, height);

  glutPostRedisplay();
}


void Draw()
{
  if (paused) return;

  cMatrix b, a, c, d;
  static DWORD clk = 0;
  static INTEG ti = 0;
  static float tf = 0;
  INTEG i, j;

// Read in from /dev/js0 and /dev/js1 for joystick input
//

  if (!js_disabled) {  
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) Fatal("/dev/js0 Read Error.\n");
    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) Fatal("/dev/js1 Read Error.\n");
  };


// Setup Planets Time (to determine SkyDome colors)
//

  planet->mDome->SetSystemDayTime();      // local time of "real" world
//  planet->mDome->SetDayTime(12.00);     // High Noon!!!
//  planet->mDome->SetDayTime(24.00);     // Mid Night!!!
  planet->mDome->SetDayTime(19.00);


// Setup to prepare to Draw
//

  cDome *dome = planet->mDome;

  if (!wireframe) {
    glClearColor(dome->mLoColor.mV[X], dome->mLoColor.mV[Y], dome->mLoColor.mV[Z], 1.0);
    glFogfv(GL_FOG_COLOR, dome->mLoColor.mV);
  } else {
    cVector dark;
    dark.Set(0.0, 0.0, 0.0, 1.0);
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glFogfv(GL_FOG_COLOR, dark.mV);
  };

  glEnable(GL_FOG);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  mainView.Use();


// ------------------------->
// - D R A W  -  B E L O W ->
// ------------------------->

glEnable(GL_CULL_FACE);

  static BOOLEAN justFired=0;
  theUser->mChasis.mV[Y] = -((kb0.x+js0.x-js0xm)/50)/1.75;
//  theUser->mTower.mV[X] = -1+sin(360*planet->getTime()*0.017454)*2;
//  theUser->mTower.mV[X] = INTEG((-(kb0.y+js0.y-js0ym)*0.01-theUser->mDevice->mTwr.mV[X]*0.20)*3.0);
//  theUser->mTower.mV[Y] = INTEG((-(js1.x-js1xm)*0.01-theUser->mDevice->mTwr.mV[Y]*0.20)*3.0);
  theUser->mThrottle = -(2*kb1.y+js1.y-js1ym)/11;
//  theUser->mDevice->Jet( ((js1.buttons & 1) || (kb1.buttons & 1)) * 100.0);
  float p = ((js1.buttons & 1) || (kb1.buttons & 1)) * 100.0f;
  theUser->mDevice->handleMsg(0, (long) "Jet", (long) p, (long) NULL);


  if ((justFired==0) && ((js0.buttons & 1) || (kb0.buttons & 1))) {
    theUser->mWeapon %= theUser->mDevice->handleMsg(0, (long) "weaponcount", 0, 0);
//    theUser->mDevice->handleMsg(0, (long) "fireall", (long) NULL, 0);
    theUser->mDevice->handleMsg(0, (long) "fire", (long) theUser->mWeapon, (long) NULL);
    theUser->mWeapon++;
//    justFired=1;
  } else if (!(js0.buttons & 1) && !(kb0.buttons & 1)) justFired=0;

  kb0.x -= kb0.x/3;
  kb0.y -= kb0.y/3;
  kb0.buttons = 0;
  kb1.buttons = 0;


// Do Animtations and To-World-Transformations.
//

	cObjectFunction* objfun = NULL;

	objfun = new cObjectFunction_animateEntity(planet->getSPF());
	planet->callObjects(objfun, 0, true);
	delete objfun;
//  planet->animateSelf(planet->getSPF());

  glPushMatrix();
  glLoadIdentity();

	objfun = new cObjectFunction_transformModel();
	planet->callObjects(objfun, 0, true);
	delete objfun;
//  planet->transformSelf();

	glPopMatrix();


// Setup camera.
//

  cMech* mech = (cMech*) theUser->mDevice;
  float *matrixfv = mech->getPilotMatrix();
  assert(matrixfv != NULL);
  glLoadIdentity();
  if (external_camera) {
    glTranslatef(0, 0, -5);
    glRotatef(15, 1, 0, 0);
  } else if (satelite_camera) {
    glTranslatef(0, 0, -100);
    glRotatef(90, 1, 0, 0);
  };
  glRotatef(180, 0, 1, 0);
  glMultMatrixf(matrixfv);
  glGetFloatv(GL_MODELVIEW_MATRIX, a.mM);


// Draw Lines from Mech No.1 to Mech No.2
//

  glDisable(GL_TEXTURE_2D);
/*
  glColor4f(1.0, 0.0, 0.0, 1.0);
  glBegin(GL_LINES);
    glVertex3fv(theAI->mSelf.mV);
    glVertex3fv(theAI->mDest.mV);
  glEnd();

  glColor4f(0.0, 1.0, 0.0, 1.0);
  glBegin(GL_LINES);
    glVertex3fv(theAI2->mSelf.mV);
    glVertex3fv(theAI2->mDest.mV);
  glEnd();
*/

// Setup the drawing scheme

  if (!wireframe) {
    glEnable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  } else {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  };

// Draw the environment: Ground, SkyDome, Clouds
//

#ifdef LIGHTON
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
#endif
  dome->DrawGround(a, TexID[7]);
#ifdef LIGHTON
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);
#endif

  if (!wireframe) {
    glDisable(GL_FOG);
    glDisable(GL_TEXTURE_2D);
    dome->DrawDome(a);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_FOG);
    dome->DrawClouds(a, TexID[3]);
  };


// Draw the things: Mechs, Buildings, Trees, Shots, ...
//

#ifdef LIGHTON
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
#endif
  glBindTexture(GL_TEXTURE_2D, TexID[6]);

  glLoadMatrixf(a.mM); // Use it!
  glColor4f(1,1,1,1);

	objfun = new cObjectFunction_drawModel();
	planet->callObjects(objfun, 0, true);
	delete objfun;
//  planet->drawSelf();

	planet->bagFragged();

#ifdef LIGHTON
  glDisable(GL_LIGHT0);
  glDisable(GL_LIGHTING);
#endif


glDisable(GL_CULL_FACE);


// Draw the 3d Consoles
//

#if 0
  b.Set();
  b.MovRel(0, 0, -1);

  con[0]->mPose.mRot.mV[Y] += sin(5*tf*0.017454)*0.1;
  con[0]->Draw(b);

  con[1]->mPose.mRot.mV[Y] -= sin(5*tf*0.017454)*0.1;
  con[1]->Draw(b);
#endif

// Make the drawn things visible and setup some counters
//

  glFlush();
  glutSwapBuffers();

  ti++;
  tf+=0.3;

  planet->updateTimers(); // calculate the time that frame took
}


void Idle()
{
  glutPostRedisplay();
}


void MouseFunc(int button, int state, int x, int y)
{
  if (button == GLUT_RIGHT_BUTTON)   rButton = state == GLUT_DOWN;
  if (button == GLUT_MIDDLE_BUTTON)  mButton = state == GLUT_DOWN;
  if (button == GLUT_LEFT_BUTTON)    lButton = state == GLUT_DOWN;
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
}


void Mouse(int x, int y)
{
  if (x > width) x = width;
  if (y > height) y = height;
  mxpos = x;
  mypos = y;
  yrSpeed = (y-height/2.)/(height/4.);
  xrSpeed = (x-width/2.)/(width/4.);
}


void Key(unsigned char k, int x, int y)
{

#ifdef USE_3DFX
  if (k == FXMODE_KEY) {
    fullscreen = 1 - fullscreen;
    XMesaSetFXmode(fullscreen ? XMESA_FX_FULLSCREEN : XMESA_FX_WINDOW);
    if (fullscreen == 0) {
      cDummy->Insert("\nWINDOWED FX MODE\n");
    } else {
      cDummy->Insert("\nFULLSCREEN FX MODE\n");
    };
    if (paused) {
      paused = 0;
      Draw();
      paused = 1;
    };
  } else
#endif

  if (k == 27) {
#ifdef USE_3DFX
    XMesaSetFXmode(XMESA_FX_WINDOW);
#endif
    exit(0);
  } else

  if (k == '1') {
    kb1.y = 0;
  } else if (k == '2') {
    kb1.y = -200;
  } else if (k == '3') {
    kb1.y = -300;
  } else if (k == '4') {
    kb1.y = -400;
  } else if (k == '5') {
    kb1.y = -500;
  } else if (k == '6') {
    kb1.y = -600;
  } else if (k == '7') {
    kb1.y = -700;
  } else if (k == '8') {
    kb1.y = -800;
  } else if (k == '9') {
    kb1.y = -900;
  } else if (k == '0') {
    kb1.y = -1000;
  } else if (k == CAMERA_KEY) {
    external_camera = 1 - external_camera;
    satelite_camera = 0;
  } else if (k == SAT_CAMERA_KEY) {
    satelite_camera = 1 - satelite_camera;
    external_camera = 0;
  } else if (k == FIRE_KEY) {
    kb0.buttons |= 1;
  } else if (k == JUMPJET_KEY) {
    kb1.buttons |= 1;
  } else if (k == WIREFRAME_KEY) {
    wireframe = 1 - wireframe;
  } else if (k == PAUSE_KEY) {
    if (!paused) {
      cDummy->Insert("\nPAUSED\n");
      Draw();
    } else {
      cDummy->Insert("\nSUSPEND\n");
    };
    paused = 1 - paused;
  } else {
    cDummy->InsertChar(k);
  };
}


void Special(int k, int x, int y)
{
  kb0.x = 0;
  if (k == GLUT_KEY_LEFT) kb0.x = -1000;
  if (k == GLUT_KEY_RIGHT) kb0.x = +1000;
  kb0.y = 0;
  if (k == GLUT_KEY_UP) kb0.y = -1000;
  if (k == GLUT_KEY_DOWN) kb0.y = +1000;
}


//
// GOD begins below!
//

#ifdef _WIN32

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  int    argc = 2;
  char  *argv[] = { "Nothing", "Nothing", "Nothing" };

#else

int main(int argc, char **argv) {

#endif

  js_disabled = 0;

  js0file = open(js0name, O_RDONLY);
  if (js0file < 0) js_disabled = 1;

  js1file = open(js1name, O_RDONLY);
  if (js1file < 0) js_disabled = 1;

  if (!js_disabled) {
    js0status = read(js0file, &js0, JS_RETURN);
    if (js0status != JS_RETURN) js_disabled = 1;

    js1status = read(js1file, &js1, JS_RETURN);
    if (js1status != JS_RETURN) js_disabled = 1;
  };
  
  js0xm = js0.x; // joystick middle is set to the current joystick position
  js0ym = js0.y;

  js1xm = js1.x;
  js1ym = js1.y;
  
  kb0.x = 0;
  kb0.y = 0;

  kb1.x = 0;
  kb1.y = 0;

  width = 640;
  height = 480;

#ifdef USE_3DFX
  putenv("MESA_GLX_FX=f");
//  putenv("MESA_GLX_FX=window");
//  putenv("SST_VGA_PASS=1");
//  putenv("SST_NOSHUTDOWN=0");
//  putenv("FX_GLIDE_NO_SPLASH=");
  width = 640;
  height = 480;
#endif

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);  
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("My Window");

#ifdef USE_3DFX
  glutFullScreen();
#endif

  glutDisplayFunc(*Draw);
  glutReshapeFunc(*Reshape);
  glutIdleFunc(*Idle);
  glutMouseFunc(*MouseFunc);
  glutPassiveMotionFunc(*Mouse);
  glutMotionFunc(*Mouse);
  glutKeyboardFunc(*Key);
  glutSpecialFunc(*Special);

  glutSetCursor(GLUT_CURSOR_NONE);
  glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);

  Init();
  glutMainLoop();

  if (!js_disabled) {  
    close(js1file);
    close(js0file);
  };

  return 0;
}

