#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <GL\glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "models.h"

#include "math3d.h"
#ifndef M_PI
#define M_PI 3.141594
#endif

#include <cassert>

using namespace std;


//-----------------------------------


static
void setupMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (pos != NULL) glTranslatef(pos[0], pos[1], pos[2]);

  if (rot != NULL) {
    glRotatef(rot[0], 1, 0, 0);
    glRotatef(rot[1], 0, 1, 0);
    glRotatef(rot[2], 0, 0, 1);
  };

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


static
void setupInvMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (rot != NULL) {
    glRotatef(-rot[2], 0, 0, 1);
    glRotatef(-rot[1], 0, 1, 0);
    glRotatef(-rot[0], 1, 0, 0);
  };
  if (pos != NULL) glTranslatef(-pos[0], -pos[1], -pos[2]);

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


//-----------------------------------


cConstruct::cConstruct()
{
  mName[0] = '\0';
  mVisible     = 1;
  mTransformed = false;
  mParent      = NULL;
  mVolume      = 0;
  mSurface     = 0;
  mDamage      = 0;
//  mPoseMatrixfv = NULL;
  mRotator3fv = NULL;
  mTranslator3fv = NULL;
}


void cConstruct::Calculate()
{
  mVolume      = 0;
  mSurface     = 0;
}


void cConstruct::ToWorld()
{
  if (mTransformed) return;
  else mTransformed = true;

  if (mParent) mParent->ToWorld();

  // Setup Inverse Matrix (Global-Coords to Local-Coords).
  glPushMatrix();
  glLoadIdentity();
//  if (mPoseInvMatrixfv != NULL) glMultMatrixf(mPoseInvMatrixfv);
  glTranslatef(-mPos.mV[X], -mPos.mV[Y], -mPos.mV[Z]);
  glRotatef(-mRot.mV[Z], 0, 0, 1);
  glRotatef(-mRot.mV[Y], 0, 1, 0);
  glRotatef(-mRot.mV[X], 1, 0, 0);
  if (mRotator3fv)
  {
    glRotatef(-mRotator3fv[Z], 0, 0, 1);
    glRotatef(-mRotator3fv[Y], 0, 1, 0);
    glRotatef(-mRotator3fv[X], 1, 0, 0);
  };
  if (mParent) glMultMatrixf(mParent->mInvMatrix.mM);
  glGetFloatv(GL_MODELVIEW_MATRIX, mInvMatrix.mM);
  glPopMatrix();


  // Setup Matrix (Local-Coords to Global-Coords).
  glPushMatrix();
  glLoadIdentity();
  if (mParent) glLoadMatrixf(mParent->mMatrix.mM);
//  if (mPoseMatrixfv != NULL) glMultMatrixf(mPoseMatrixfv);
  if (mRotator3fv)
  {
    glRotatef(mRotator3fv[X], 1, 0, 0);
    glRotatef(mRotator3fv[Y], 0, 1, 0);
    glRotatef(mRotator3fv[Z], 0, 0, 1);
  };
  glRotatef(mRot.mV[X], 1, 0, 0);
  glRotatef(mRot.mV[Y], 0, 1, 0);
  glRotatef(mRot.mV[Z], 0, 0, 1);
  glTranslatef(mPos.mV[X], mPos.mV[Y], mPos.mV[Z]);
  glGetFloatv(GL_MODELVIEW_MATRIX, mMatrix.mM);
  glPopMatrix();

//  mSphere.mV.Set(0, 0, 0);
}


void cConstruct::Draw()
{
  return;
  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible)
  {
  };

  glPopMatrix();
}


bool cConstruct::collideLine(float* a, float* b, float* point, float* normal, float* dist)
{
  cVector a_, b_;
  a_.Set(a[0], a[1], a[2]);
  b_.Set(b[0], b[1], b[2]);

  cMatrix im;
  memcpy(im.mM, mInvMatrix.mM, sizeof(float) * 16);
//  im.Apply(a_);
//  im.Apply(b_);

  cRay ray;
  ray.Set(a_, b_);

  cVector pnt;
  if (mSphere.Hit(ray, pnt))
  {
    cVector delta;
    delta.Set(b_, a_);
    delta.Normalise();
         if (fabs(delta.mV[0]) > 0.1) *dist = (pnt.mV[0] - a_.mV[0]) /  delta.mV[0];
    else if (fabs(delta.mV[1]) > 0.1) *dist = (pnt.mV[1] - a_.mV[1]) /  delta.mV[1];
    else if (fabs(delta.mV[2]) > 0.1) *dist = (pnt.mV[2] - a_.mV[2]) /  delta.mV[2];

    mDamage += 0.01f;
//    if (dist >= 0)
    {
      normal[0] = -delta.mV[0];
      normal[1] = -delta.mV[1];
      normal[2] = -delta.mV[2];
			point[0] = pnt.mV[0];
			point[1] = pnt.mV[1];
			point[2] = pnt.mV[2];
      return true;
    };
  };

  return false;
}


//-----------------------------------


void cCriss::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCriss::Draw()
{
  int   i;
  float W2 = mWidth/2.0;
  float H2 = mHeight/2.0;
  float L2 = mLength/2.0;
  cVector p[8];

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {

    p[0].Set(    -W2,     +H2,      0);
    p[1].Set(    -W2,     -H2,      0);
    p[2].Set(      0,     +H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    +W2,     +H2,      0);
    p[5].Set(    +W2,     -H2,      0);
    p[6].Set(      0,     +H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    float wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);
    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[2].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[1].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[1].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[5].mV);

    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


void cCross::Calculate()
{
  mVolume = mWidth*mHeight*mLength+0.001;
  mSurface = 1.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCross::Draw()
{
  int   i;
  float W2 = mWidth/2.0;
  float H2 = mHeight/2.0;
  float L2 = mLength/2.0;
  cVector p[12];

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {

    p[0].Set(    -W2,       0,    -L2);
    p[1].Set(    +W2,       0,    -L2);
    p[2].Set(      0,      H2,    -L2);
    p[3].Set(      0,     -H2,    -L2);

    p[4].Set(    -W2,       0,    +L2);
    p[5].Set(    +W2,       0,    +L2);
    p[6].Set(      0,      H2,    +L2);
    p[7].Set(      0,     -H2,    +L2);

    p[8].Set(    -W2,     -H2,      0);
    p[9].Set(    +W2,     -H2,      0);
    p[10].Set(   +W2,     +H2,      0);
    p[11].Set(   -W2,     +H2,      0);

    float wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);

    glBegin(GL_QUADS);

    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[4].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[5].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[7].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[6].mV);

    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[9].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[10].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[11].mV);

    glTexCoord2f(1, 0);
    glVertex3fv(p[9].mV);
    glTexCoord2f(0, 0);
    glVertex3fv(p[8].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[11].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[10].mV);

    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


void cQuader::Calculate()
{
  mVolume = mWidth*mHeight*mLength;
  mSurface = 2.0*(mWidth*mHeight + mWidth*mLength + mHeight*mLength);
  mSphere.mR = sqrt(xsqr(mWidth/2.0) + xsqr(mHeight/2.0) + xsqr(mLength/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cQuader::Draw()
{
  int   i;
  float W2 = mWidth/2.0;
  float H2 = mHeight/2.0;
  float L2 = mLength/2.0;
  cVector p[8];

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {

    p[0].Set(     0,       0,      0);
    p[1].Set(mWidth,       0,      0);
    p[2].Set(mWidth, mHeight,      0);
    p[3].Set(     0, mHeight,      0);
    p[4].Set(     0,       0, mLength);
    p[5].Set(mWidth,       0, mLength);
    p[6].Set(mWidth, mHeight, mLength);
    p[7].Set(     0, mHeight, mLength);

    for (i=0; i<8; i++)
    {
      p[i].mV[X] -= W2;
      p[i].mV[Y] -= H2;
      p[i].mV[Z] -= L2;
    };

    float wm = 1.0-mDamage/mVolume;
    if (wm<0) wm = 0;
    if (wm>1) wm = 1;
    glColor4f(1.0, wm, wm, 1.0);

    glBegin(GL_QUADS);

    glNormal3f(0, 0, -1);
    glTexCoord2f(0, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[3].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[2].mV);

    glNormal3f(0, 0, 1);
    glTexCoord2f(0, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[5].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[6].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[7].mV);

    glNormal3f(0, -1, 0);
    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[5].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[4].mV);

    glNormal3f(0, 1, 0);
    glTexCoord2f(0, 0);
    glVertex3fv(p[2].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[3].mV);
    glTexCoord2f(1, 1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(0, 1);
    glVertex3fv(p[6].mV);

    glNormal3f(1, 0, 0);
    glTexCoord2f(0, 0);
    glVertex3fv(p[0].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[4].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[7].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[3].mV);

    glNormal3f(-1, 0, 0);
    glTexCoord2f(0, 0);
    glVertex3fv(p[5].mV);
    glTexCoord2f(1, 0);
    glVertex3fv(p[1].mV);
    glTexCoord2f(1, -1);
    glVertex3fv(p[2].mV);
    glTexCoord2f(0, -1);
    glVertex3fv(p[6].mV);

    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


void cCylinder::Calculate()
{
  mVolume = M_PI * xsqr(mRadius) * mHeight;
  mSurface = 2.0 * M_PI * mRadius * (mRadius * mHeight);
  mSphere.mR = xmax(mRadius, (mHeight/2.0));
  mSphere.mR2 = xsqr(mSphere.mR);
}


void cCylinder::Draw()
{
  int   i;
  float wm = 1.0-mDamage/mVolume;
  if (wm<0) wm = 0;
  if (wm>1) wm = 1;
  glColor4f(1.0, wm, wm, 1.0);

  glPushMatrix();
  glMultMatrixf(mMatrix.mM);

  if (mVisible) {
    glBegin(GL_TRIANGLE_STRIP);
    for (i=0; i<=8; i++) {
      glTexCoord2f(i/8.0, 1);
      glVertex3f(sin(i*45.0*0.017454)*mRadius, +mHeight/2.0, cos(i*45.0*0.017454)*mRadius);
      glTexCoord2f(i/8.0, 0);
      glVertex3f(sin(i*45.0*0.017454)*mRadius, -mHeight/2.0, cos(i*45.0*0.017454)*mRadius);
    };
    glEnd();
  };

  glPopMatrix();
}


//-----------------------------------


/*
void cPlantModel_DrawPart(int depth, int maxdepth, float lenght)
{
  if (depth >= maxdepth) return;

  const int n = 3;
  const float height = lenght*(maxdepth - depth);
  const float width = 0.2*(maxdepth - depth);
//  const float width = 1;
  
  glPushMatrix();
  glTranslatef(0, height*0.5f, 0);
  glScalef(width, height, width);
//  glScalef(1, 3, 1);
  glColor4f(0.2f, 0.2f, 0.1f, 1.0f);
  glutSolidCube(1);
  glPopMatrix();

  int i;
  for (i = 0; i < n; i++)
  {
    glPushMatrix();

    glTranslatef(0, (i + 2)*height/float(n + 1), 0);
    glRotatef(i*360/float(n), 0, 1, 0);
    glRotatef(35, 1, 0, 0);
    cPlantModel_DrawPart(depth+1, maxdepth, lenght);

    glPopMatrix();
  };
}
*/

void cPlantModel_DrawPart(int depth, int maxdepth, float length, list<int>::iterator &it)
{
  if (depth >= maxdepth) return;

  const float height = length*(maxdepth - depth);
  const float width = 0.035*length+(maxdepth - depth)*(maxdepth - depth)*length*0.0125f;

//  bool leaves = false;
  bool leaves = true;
  if ((depth == maxdepth-1) && leaves) {
    glPushMatrix();

    glTranslatef(0, height*0.5f, 0);
    glScalef(width, height, width);
    glutSolidCube(1);

    glTranslatef(0, 0.5f, 0);
    glScalef(height/width/2.0f, 0.25, height/width/2.0f);
//    glScalef(height/3, height/2, height/3);
//    glColor3f(0.8f, 0.7f, 0.0f);
    glColor3f(0.0f, 0.4f, 0.0f);
    glutSolidSphere(1, 3, 2);
    glColor3f(0.1f, 0.1f, 0.0f);

    glPopMatrix();
  } else {
    glPushMatrix();
    glTranslatef(0, height*0.5f, 0);
    glScalef(width, height, width);
    glutSolidCube(1);
    glPopMatrix();
  };


  int anglex, angley;
  const int n = 3;
  for (int i = 0; i < n; i++)
  {
    glPushMatrix();

    angley = *it; it++;
    anglex = *it; it++;

    glTranslatef(0, (i + 2)*height/float(n + 1)-width*0.5f, 0);
    glRotatef(angley, 0, 1, 0);
    glRotatef(anglex, 1, 0, 0);
    cPlantModel_DrawPart(depth+1, maxdepth, length, it);

    glPopMatrix();
  };
}


void cPlantModel_GeneratePart(int depth, int maxdepth, list<int> &values)
{
  if (depth >= maxdepth) return;

  const int yvar = 15;
  const float xvar = 27.0f;
  const int n = 3;
  for (int i = 0; i < n; i++)
  {
    // angley
    values.push_back(int(i*360.0f/float(n)) - yvar + rand()%(2*yvar) );
    // anglex
    values.push_back(int(60.0f - xvar + (rand() % 100)*0.01f * (2.0f*xvar)*(1.0f-float(maxdepth-depth)/float(maxdepth))) );
    cPlantModel_GeneratePart(depth+1, maxdepth, values);
  };
}


void* cPlantModel::createContext(char* descriptor)
{
  context* co = new context;

  co->rotator3fv = NULL;
  co->translator3fv = NULL;
  setupMatrix(co->matrix);
  setupInvMatrix(co->invmatrix);

  int d = 2;
  float h = 3.0f;

  if (descriptor)
  {
    sscanf(descriptor, "%i %f", &d, &h);
  };

  co->displaylist = -1;
  int depth = d;
  co->values.push_back(depth);
  int height = int(h * 100.0f);
  co->values.push_back(height);

  cPlantModel_GeneratePart(0, depth, co->values);
  return co;
}


void cPlantModel::transformSelf(cEntityModel *entmod)
{
  context *co;
  co = (context*) entmod->mContext;

  setupInvMatrix(co->invmatrix, co->rotator3fv, co->translator3fv);
  setupMatrix(co->matrix, co->rotator3fv, co->translator3fv);
}


void cPlantModel::drawSelf(cEntityModel *entmod)
{
  context *co;
  co = (context*) entmod->mContext;

  glDisable(GL_TEXTURE_2D);

  glPushMatrix();

  glMultMatrixf(co->matrix);

  if (co->displaylist < 0)
  {
    co->displaylist = glGenLists(1);
    glNewList(co->displaylist, GL_COMPILE_AND_EXECUTE);
    glEnable(GL_NORMALIZE);

    list<int>::iterator it = co->values.begin();
    int depth = *it; it++;
    float height = *it; it++;
    height *= 0.01;

    glPushMatrix();
    glScalef(depth*depth*height*0.16f, 0, depth*depth*height*0.16f);
    glColor4f(0.0f, 0.0f, 0.0f, 0.2);
    glutSolidSphere(1, 4, 5);
    glColor3f(0.1f, 0.1f, 0.0f);
    glPopMatrix();

    glColor3f(0.1f, 0.1f, 0.0f);
    cPlantModel_DrawPart(0, depth, height, it);

    glDisable(GL_NORMALIZE);
    glEndList();
  } else {
    glCallList(co->displaylist);
  };

  glPopMatrix();
}


void cPlantModel::setRotator(cEntityModel* entmod, float* rot3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  co->rotator3fv = rot3fv;
}


void cPlantModel::setTranslator(cEntityModel* entmod, float* tra3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  co->translator3fv = tra3fv;
}


//--------------------------------


#include <map>

bool cConstructModel::model::loadSelf(FILE *f)
{
  int           nparts, s, t;
  cCylinder    *cy;
  cQuader      *q;
  cCriss       *ci;
  cCross       *cr;
  float        x, y, z, a, b, c;

  typedef map<string, int> symmap;
  symmap conmap;
  conmap.insert(symmap::value_type("construct", cConstruct::Construct));
  conmap.insert(symmap::value_type("cylinder", cConstruct::Cylinder));
  conmap.insert(symmap::value_type("quader", cConstruct::Quader));
  conmap.insert(symmap::value_type("criss", cConstruct::Criss));
  conmap.insert(symmap::value_type("cross", cConstruct::Cross));

  char line[256];
  char par[64], nam[64], con[64];

  fscanf(f, "%ld %ld\n", &nparts, &s);

  cConstruct* construct;
  for (int k=0; k<nparts; k++)
  {
    fgets(line, 256, f);

    string contype_s;

    sscanf(line, "%s %s %s   %f %f %f", par, nam, con, &x, &y, &z);

    switch (conmap[con]) {

      case cConstruct::Construct: {
        construct = new cConstruct;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Construct;
        construct->mPos.Set(x, y, z);
      } break;
      case cConstruct::Cylinder: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f", par, nam, con,  &x, &y, &z,  &a, &b);
        construct = new cCylinder;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Cylinder;
        construct->mPos.Set(x, y, z);
        cy = (cCylinder*) construct;
        cy->mRadius = a;
        cy->mHeight = b;
        construct->Calculate();
      } break;
      case cConstruct::Quader: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cQuader;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Quader;
        construct->mPos.Set(x, y, z);
        q = (cQuader*) construct;
        q->mWidth = a;
        q->mHeight = b;
        q->mLength = c;
        construct->Calculate();
      } break;
      case cConstruct::Criss: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cCriss;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Criss;
        construct->mPos.Set(x, y, z);
        ci = (cCriss*) construct;
        ci->mWidth = a;
        ci->mHeight = b;
        ci->mLength = c;
        construct->Calculate();
      } break;
      case cConstruct::Cross: {
        sscanf(line, "%s %s %s   %f %f %f   %f %f %f", par, nam, con,  &x, &y, &z,  &a, &b, &c);
        construct = new cCross;
        strcpy(construct->mParentName, par);
        strcpy(construct->mName, nam);
        construct->mType = cConstruct::Cross;
        construct->mPos.Set(x, y, z);
        cr = (cCross*) construct;
        cr->mWidth = a;
        cr->mHeight = b;
        cr->mLength = c;
        construct->Calculate();
      } break;
      default: assert("No such construct" == NULL); break;
    };
    constructs.push_back(construct);
  };

  cConstruct *construct_i;
  cConstruct *construct_j;
  list<cConstruct*>::iterator i;
  list<cConstruct*>::iterator j;

  for (i = constructs.begin(); i != constructs.end();)
  {
    construct_i = *i; i++;
    construct_i->mParent = NULL; //construct;
    if (strcmp(construct_i->mParentName, "NULL") == 0) continue;
    for (j = constructs.begin(); j != constructs.end();)
    {
      construct_j = *j; j++;
      if (construct_i == construct_j) continue;
      if (strcmp(construct_i->mParentName, construct_j->mName) == 0)
      {
        construct_i->mParent = construct_j;
        break;
      };
    };
  };

  return false;
}


void* cConstructModel::createContext(char* descriptor)
{
  context* co = new context;
  co->mod = NULL;
  co->rotator3fv = NULL;
  co->translator3fv = NULL;
  setupMatrix(co->matrix);
  setupInvMatrix(co->invmatrix);
  
  if (descriptor == NULL) return co;

  string desc = descriptor;
  for (list<model*>::iterator mit = mModels.begin(); mit != mModels.end();)
  {
    model* mod = *mit; mit++;
    if (mod->name == desc) {
      context* co = new context;
      co->mod = mod;
      for (int i = 0; i < mod->constructs.size(); i++)
      {
        co->rotators.push_back(NULL);
        co->translators.push_back(NULL);
        co->matrices.push_back(new float[16]);
        co->invmatrices.push_back(new float[16]);
      };
      return co;
    };
  };

  // If there is no such model then try to load it.

  FILE* f;
  f = fopen(descriptor, "rt");
  // If Model not available then return NULL.
  if (f == NULL) return NULL;
  model* mod = new model;
  if (mod->loadSelf(f)) {
    assert("Couldn't load construct" == NULL);
    delete mod;
    return NULL;
  };
  co->mod = mod;
  for (int i = 0; i < mod->constructs.size(); i++)
  {
    co->rotators.push_back(NULL);
    co->translators.push_back(NULL);
    co->matrices.push_back(new float[16]);
    co->invmatrices.push_back(new float[16]);
  };
  mModels.push_back(mod);
  fclose(f);
  return co;
}


void cConstructModel::transformSelf(cEntityModel *entmod)
{
  if (entmod->mContext == NULL) return;
  context* co = (context*) entmod->mContext;

  setupInvMatrix(co->invmatrix, co->rotator3fv, co->translator3fv);
  setupMatrix(co->matrix, co->rotator3fv, co->translator3fv);

  cConstruct *construct;
  list<cConstruct*>::iterator i;
  float* rot3fv;
  float* tra3fv;
  list<float*>::iterator r;
  list<float*>::iterator t;
  for (i = co->mod->constructs.begin(), r = co->rotators.begin(), t = co->translators.begin(); i != co->mod->constructs.end();)
  {
    rot3fv = *r; r++;
    tra3fv = *t; t++;
    construct = *i; i++;
    construct->mRotator3fv = rot3fv;
    construct->mTranslator3fv = tra3fv;
    construct->mTransformed = false;
  };

  list<float*>::iterator j;
  list<float*>::iterator k;
  float* mat;
  for (i = co->mod->constructs.begin(), j = co->matrices.begin(), k = co->invmatrices.begin(); i != co->mod->constructs.end();)
  {
    construct = *i; i++;

    construct->ToWorld();

    mat = *j; j++;
    memcpy(mat, construct->mMatrix.mM, sizeof(float) * 16);

    mat = *k; k++;
    memcpy(mat, construct->mInvMatrix.mM, sizeof(float) * 16);
  };
}


void cConstructModel::drawSelf(cEntityModel *entmod)
{
  glPushMatrix();

  if (entmod->mContext == NULL) return;
  context* co = (context*) entmod->mContext;

  // Put model to required position.
  glMultMatrixf(co->matrix);

/*
  glDisable(GL_TEXTURE_2D);
  glColor4f(0.6, 0.6, 0.6, 0.4);
  glutSolidSphere(1.4f, 3, 3);
  glEnable(GL_TEXTURE_2D);
*/

//  glDisable(GL_TEXTURE_2D);
  glEnable(GL_TEXTURE_2D);
  cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  float* mat;
  for (i = co->mod->constructs.begin(), j = co->matrices.begin(); i != co->mod->constructs.end();)
  {
    construct = *i; i++;
    mat = *j; j++;
    memcpy(construct->mMatrix.mM, mat, sizeof(float) * 16);
    construct->Draw();
  };
  glEnable(GL_TEXTURE_2D);

  glPopMatrix();
}


cConstruct* cConstructModel::findConstruct(cEntityModel* entmod, char* name, list<float*>::iterator &it)
{
  context *co;
  co = (context*) entmod->mContext;

  cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  for (i = co->mod->constructs.begin(); i != co->mod->constructs.end();)
  {
    construct = *i; i++;
    if (strcmp(construct->mName, name) == 0) return construct;
    it++;
  };
  return NULL;
}


void cConstructModel::setRotator(cEntityModel* entmod, float* rot3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  co->rotator3fv = rot3fv;
}


void cConstructModel::setTranslator(cEntityModel* entmod, float* tra3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  co->translator3fv = tra3fv;
}


void cConstructModel::setConstructRotator(cEntityModel* entmod, char* name, float* rot3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  list<float*>::iterator i = co->rotators.begin();
  if (findConstruct(entmod, name, i))
  {
    *i = rot3fv;
  };
}


void cConstructModel::setConstructTranslator(cEntityModel* entmod, char* name, float* tra3fv)
{
  context *co;
  co = (context*) entmod->mContext;
  list<float*>::iterator i = co->translators.begin();
  if (findConstruct(entmod, name, i))
  {
    *i = tra3fv;
  };
}


float* cConstructModel::getMatrix(cEntityModel* entmod)
{
  context *co;
  co = (context*) entmod->mContext;
  return co->matrix;
}


float* cConstructModel::getInvMatrix(cEntityModel* entmod)
{
  context *co;
  co = (context*) entmod->mContext;
  return co->invmatrix;
}


float* cConstructModel::getConstructMatrix(cEntityModel* entmod, char* name)
{
  context *co;
  co = (context*) entmod->mContext;
  list<float*>::iterator i = co->matrices.begin();
  if (findConstruct(entmod, name, i))
  {
    float* matfv = *i;
    return matfv;
  };
  return NULL;
}


float* cConstructModel::getConstructInvMatrix(cEntityModel* entmod, char* name)
{
  context *co;
  co = (context*) entmod->mContext;
  list<float*>::iterator i = co->invmatrices.begin();
  if (findConstruct(entmod, name, i))
  {
    float* matfv = *i;
    return matfv;
  };
  return NULL;
}

/*
void cConstructModel::collideLine(cEntityModel *entmod, float* a, float* b, cModelCollision *&result)
{
  context *co;
  co = (context*) entmod->mContext;

	// Transform line into local coordinate system using the inverse matrix from the context.
	cMatrix m;
	memcpy(m.mM, co->invmatrix, sizeof(float) * 16);
	cVector a_, b_;
	a_.Set(a[0], a[1], a[2]);
	b_.Set(b[0], b[1], b[2]);
	m.Apply(a_);
	m.Apply(b_);

	cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  float* mat;
	float normal[3], point[3], dist;
  for (i = co->mod->constructs.begin(), j = co->invmatrices.begin(); i != co->mod->constructs.end();)
  {
    construct = *i; i++;
    mat = *j; j++;
    memcpy(construct->mInvMatrix.mM, mat, sizeof(float) * 16);
    construct->collideLine(a_.mV, b_.mV, normal, point, &dist);
//    construct->collideLine(a, b, normal, point, &dist);
  };
}
*/


void cConstructModel::collideLine(cEntityModel *entmod, float* a, float* b, cModelCollision *&result)
{
  context *co;
  co = (context*) entmod->mContext;

	// Transform line into local coordinate system using the inverse matrix from the context.
	cMatrix m, n;
	memcpy(m.mM, co->invmatrix, sizeof(float) * 16);
	cVector a_, b_;

	cConstruct *construct;
  list<cConstruct*>::iterator i;
  list<float*>::iterator j;
  float* mat;
	float normal[3], point[3], dist;
  for (i = co->mod->constructs.begin(), j = co->invmatrices.begin(); i != co->mod->constructs.end();)
  {
    construct = *i; i++;
    mat = *j; j++;
    memcpy(construct->mInvMatrix.mM, mat, sizeof(float) * 16);
	  memcpy(n.mM, mat, sizeof(float) * 16);

		cMatrix o;
		o = m;
		o.Mul(n);

		a_.Set(a[0], a[1], a[2]);
  	b_.Set(b[0], b[1], b[2]);

	  o.Apply(a_);
	  o.Apply(b_);

    construct->collideLine(a_.mV, b_.mV, normal, point, &dist);
//    construct->collideLine(a, b, normal, point, &dist);
  };
}


//--------------------------------


void* cDummyModel::createContext(char* descriptor)
{
  context* co = new context;
  co->matrixfv = NULL;
  return co;
}


void cDummyModel::transformSelf(cEntityModel *entmod)
{
}


void cDummyModel::drawSelf(cEntityModel *entmod)
{
  if (entmod->mContext == NULL) return;
  context* co = (context*) entmod->mContext;
  if (co->matrixfv == NULL) return;

  glMultMatrixf(co->matrixfv);
  glDisable(GL_TEXTURE_2D);
//  glColor4f(0.6, 0.6, 0.6, 0.4);
//  glutSolidSphere(1.4f, 3, 3);
  glEnable(GL_TEXTURE_2D);
}


void cDummyModel::setPlacementMatrix(cEntityModel* entmod, float* matrixfv)
{
  context *co;
  co = (context*) entmod->mContext;
  co->matrixfv = matrixfv;
}


//--------------------------------


void* cParticleModel::createContext(char* descriptor)
{
  cParticleModel::particlesystem *ps = new cParticleModel::particlesystem;
  ps->sourceMatrixfv = NULL;
  ps->countdown = 100.0f;
  ps->v[0] = 0;
  ps->v[1] = 0;
  ps->v[2] = 1.75;
  ps->v_variation[0] = 0.1;
  ps->v_variation[1] = 0.1;
  ps->v_variation[2] = 0.0;
  return ps;
}


void cParticleModel::transformSelf(cEntityModel *entmod)
{
  // Should be animateSelf(cEntityModel *entmod);
  cParticleModel::particlesystem *ps;
  ps = (cParticleModel::particlesystem*) entmod->mContext;

  particle* pt;
  list<particle*>::iterator it;
  for (it = ps->particles.begin(); it != ps->particles.end();)
  {
    pt = *it; it++;
    pt->p[0] += pt->v[0];
    pt->p[1] += pt->v[1];
    pt->p[2] += pt->v[2];
  };
}


void cParticleModel::drawSelf(cEntityModel *entmod)
{
  cParticleModel::particlesystem *ps;
  ps = (cParticleModel::particlesystem*) entmod->mContext;

  particle* pt;
  list<particle*>::iterator it;
//  glDisable(GL_TEXTURE_2D);
  glColor4f(0.75, 0.8, 0.3, 1.0f);
  for (it = ps->particles.begin(); it != ps->particles.end();)
  {
    pt = *it; it++;
    glBegin(GL_LINES);
    glNormal3f(0, 1, 0);
    glVertex3fv(pt->p);
    glNormal3f(0, 1, 0);
    glVertex3f(pt->p[0]-pt->v[0], pt->p[1]-pt->v[1], pt->p[2]-pt->v[2]);
    glEnd();
  };
//  glEnable(GL_TEXTURE_2D);
}


void cParticleModel::setSourceMatrix(cEntityModel* entmod, float* matrixfv)
{
  cParticleModel::particlesystem *ps;
  ps = (cParticleModel::particlesystem*) entmod->mContext;
  ps->sourceMatrixfv = matrixfv;
}


void cParticleModel::setCountdown(cEntityModel* entmod, float countdown)
{
  cParticleModel::particlesystem *ps;
  ps = (cParticleModel::particlesystem*) entmod->mContext;
  ps->countdown = countdown;
}


void cParticleModel::spawnParticle(cEntityModel* entmod)
{
  cParticleModel::particlesystem *ps;
  ps = (cParticleModel::particlesystem*) entmod->mContext;

  cParticleModel::particle *pt;
  pt = new cParticleModel::particle;
  pt->countdown = ps->countdown;

  cMatrix m;
  memcpy(m.mM, ps->sourceMatrixfv, sizeof(float) * 16);
  cVector p;
  m.Apply(p);
  memcpy(pt->p, p.mV, sizeof(float)*3);

  cVector v;
  v.Set(ps->v[0], ps->v[1], ps->v[2]);
  m.Apply(v);
  v.Sub(p);
  memcpy(pt->v, v.mV, sizeof(float)*3);

  ps->particles.push_front(pt);
}


//--------------------------------

