#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "objects.h"

#include "math3d.h"
#ifndef M_PI
#define M_PI 3.141594
#endif

#include "envir.h"

#undef xmod
#undef xabs

#include <time.h>
#include <string.h>
#include <cassert>

using namespace std;


//-----------------------------------


static
void setupMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (pos != NULL) glTranslatef(pos[0], pos[1], pos[2]);

  if (rot != NULL) {
    glRotatef(rot[0], 1, 0, 0);
    glRotatef(rot[1], 0, 1, 0);
    glRotatef(rot[2], 0, 0, 1);
  };

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


static
void setupInvMatrix(float *mat, float *rot = NULL, float *pos = NULL)
{
  glPushMatrix();
  glLoadIdentity();

  if (rot != NULL) {
    glRotatef(-rot[2], 0, 0, 1);
    glRotatef(-rot[1], 0, 1, 0);
    glRotatef(-rot[0], 1, 0, 0);
  };
  if (pos != NULL) glTranslatef(-pos[0], -pos[1], -pos[2]);

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


//-----------------------------------


cWorld::cWorld()
{
/*
  mWind[X] = 0;
  mWind[Y] = 0;
  mWind[Z] = 0;
*/
  mGravity[X] = 0;
  mGravity[Y] = 9.8f;
  mGravity[Z] = 0;

  mBufferedCorpses = &mCorpses[0];

  updateTimers();
  updateTimers(); // yes, two times!!
}

#define MMTIMER

#if defined(_WIN32) && defined(MMTIMER)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#else
#include <sys/timeb.h>
#endif

float GetSPF(unsigned long *newClock, unsigned long *oldClock)
{
  float spf;

#if defined(_WIN32) && defined(MMTIMER)
  // win only
  int ok;
  ok = 0;
  if (timeBeginPeriod(1) == TIMERR_NOERROR) ok = 1;
  newClock[0] = timeGetTime();
  if (ok) timeEndPeriod(1);
  spf = (float) (newClock[0] - oldClock[0]) / 1000.0f;
#else
  // Possibly this will work on every system?
  struct timeb timb;
  ftime(&timb);
  newClock[0] = timb.millitm + 1000 * timb.time;
  spf = (float) ((double) (newClock[0] - oldClock[0]) / 1000.0);
/*
  // Works everywhere - but only gets process time (sux on win).
  newClock[0] = clock();
  spf = (float) (newClock[0] - oldClock[0]) / CLOCKS_PER_SEC;
*/
#endif
  oldClock[0] = newClock[0];
  return spf;
}


void cWorld::updateTimers()
{
  mSPF = GetSPF(&mNewClock, &mOldClock);
  mFPS = 1.0 / (mSPF + 0.000001f);
  mTime += mSPF;
  /*
  mNewClock = clock();
  mSPF = float(mNewClock - mOldClock) / CLOCKS_PER_SEC;
  mFPS = 1.0 / (mSPF + 0.000001f);
  mOldClock = mNewClock;
  */
}


void cWorld::spawnObject(cObject *object)
{
  object->setWorld(this);
//  mObjects.push_front(object);
  mObjects.push_back(object);
  object->onSpawn();
}


void cWorld::fragObject(cObject *object)
{
  if (object->getWorld() != this) return;
  if (object->getFragged()) return;
  object->setFragged();
  mBufferedCorpses->push_back(object);
  object->onFrag();
}

/*
void cWorld::animateSelf(float spf)
{
  list<cObject *>::iterator it;
  cObject *object;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
//    if (object->getController()) object->getController()->processSelf();
    object->animateSelf(spf);
  };
}


void cWorld::transformSelf()
{
  list<cObject *>::iterator it;
  cObject *object;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    glPushMatrix();
    object->transformSelf();
    glPopMatrix();
  };
}


void cWorld::drawSelf()
{
  list<cObject *>::iterator it;
  cObject *object;
  list<cObject *> visgrps[16];
  int grp;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
//    if (object->getFragged()) continue;
    grp = object->getCallGroup();
    if (grp < 0) object->setCallGroup(0);
    if (grp > MAX_VISGRP) object->setCallGroup(MAX_VISGRP - 1);
    visgrps[grp].push_back(object);
//    glPushMatrix();
//    object->Draw();
//    glPopMatrix();
  };

  for (int i = 15; i >= 0; i--)
  {
    for (it = visgrps[i].begin(); it != visgrps[i].end();)
    {
      object = *it; it++;
      if (object->getFragged()) continue;
      glPushMatrix();
      object->drawSelf();
      glPopMatrix();
    };
    visgrps[i].clear();
  };
};
*/

void cWorld::bagFragged()
{
  // Switch corpse-double-buffer.
  if (mBufferedCorpses == &mCorpses[0])
  {
    mBufferedCorpses = &mCorpses[1];
  } else {
    mBufferedCorpses = &mCorpses[0];
  };

  // Finally delete objects which where scheduled to be
  // deleted in the last frame.
  while (!mBufferedCorpses->empty())
  {
    cObject* object = mBufferedCorpses->front();
    mBufferedCorpses->pop_front();
    mObjects.remove(object);
    delete object;
  };
}

/*
bool cWorld::scanName(string name, cObject *&objectWithName, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectWithName = NULL;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    if (object->getName() == name)
    {
      objectWithName = object;
      return true;
    };
  };
  return false;
}


bool cWorld::scanName(string name, list<cObject *> *objectsWithName, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectsWithName->clear();
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    if (object->getName() == name)
    {
      objectsWithName->push_back(object);
    };
  };
  return (!objectsWithName->empty());
}
*/

/*
bool cWorld::scanRange(float center[3], float radius, cObject *&objectWithinRange, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectWithinRange = NULL;
  float dist;
  float min_radius = radius;
  cVector v_center;
  v_center.mV[0] = center[0];
  v_center.mV[1] = center[1];
  v_center.mV[2] = center[2];
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      dist = lineDistance(v_center.mV, object->mPos.mV);
      if (dist <= min_radius)
      {
        min_radius = dist;
        objectWithinRange = object;
      };
    };
  };
  return (objectWithinRange != NULL);
  return false;
}
*/

/*
bool cWorld::scanRange(float center[3], float radius, list<cObject *> *objectsWithinRange, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectsWithinRange->clear();
  float dist;
  float min_radius = radius;
  cVector v_center;
  v_center.mV[0] = center[0];
  v_center.mV[1] = center[1];
  v_center.mV[2] = center[2];
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      dist = lineDistance(v_center.mV, object->mPos.mV);
      if (dist <= min_radius)
      {
        objectsWithinRange->push_back(object);
      };
    };
  };
  return (!objectsWithinRange->empty());
  return false;
}
*/

/*
bool cWorld::scanSphere(float center[3], float radius, cObject *&objectWithinSphere, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectWithinSphere = NULL;
  float dist;
  float min_radius = radius;
  cSphere sphere;
  sphere.mR = radius;
  sphere.mR2 = sphere.mR * sphere.mR;
  sphere.mV.mV[0] = center[0];
  sphere.mV.mV[1] = center[1];
  sphere.mV.mV[2] = center[2];
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      if (object->scanSphere(sphere))
      {
        objectWithinSphere = object;
        return true;
      };
    };
  };
  return false;
}


bool cWorld::scanSphere(float center[3], float radius, list<cObject *> *objectsWithinSphere, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  objectsWithinSphere->clear();
  float dist;
  float min_radius = radius;
  cSphere sphere;
  sphere.mR = radius;
  sphere.mR2 = sphere.mR * sphere.mR;
  sphere.mV.mV[0] = center[0];
  sphere.mV.mV[1] = center[1];
  sphere.mV.mV[2] = center[2];
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      if (object->scanSphere(sphere))
      {
        objectsWithinSphere->push_back(object);
      };
    };
  };
  return (!objectsWithinSphere->empty());
}
*/

/*
bool cWorld::callObjects(cObject *excluded, int msg, long prma, long prmb, long prmc, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object == excluded) continue;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      object->handleMsg(msg, prma, prmb, prmc);
    };
  };
  return true;
}


bool cWorld::collectObjects(list<cObject*> &returncollection, int limitsize, cObject *excluded, int msg, long prma, long prmb, long prmc, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  int i = 0;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (i == limitsize) break;
    if (object == excluded) continue;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      if (object->handleMsg(msg, prma, prmb, prmc) == -1)
      {
        returncollection.push_back(object);
        i++;
      };
    };
  };
  return true;
}
*/

bool cWorld::callObjects(cObjectFunction* objfun, int groupMask, bool matchWholeMask)
{
  cObject *object;
  list<cObject *>::iterator it;
  for (it = mObjects.begin(); it != mObjects.end();)
  {
    object = *it; it++;
    if (object->getFragged()) continue;
    if ((((object->getGroupFlags() & groupMask) == groupMask) &&   matchWholeMask) ||
        (((object->getGroupFlags() & groupMask) !=         0) && (!matchWholeMask)))
    {
      objfun->applyFunction(object);
			if (!objfun->moreObjects()) break;
    };
  };
  return true;
}


//--------------------------------


cObject::cObject()
{
  mName         = "Anonymous";
  mWorld        = NULL;
  mFragged      = false;

  mGroupFlags   = 0;
  
  mCallGroup    = cWorld::GIGA;
}


void cObject::copySelf(cObject *source)
{
  mName         = source->mName;
  mWorld        = NULL;
  mFragged      = false;

  mGroupFlags   = source->mGroupFlags;

  mCallGroup    = source->mCallGroup;
}


void cObject::onSpawn()
{
}


void cObject::onFrag()
{
}


int cObject::handleMsg(int msg, long prma, long prmb, long prmc)
{
  return 0;
}


//--------------------------------


cModel::cModel()
{
}


void* cModel::createContext(char* descriptor)
{
  return NULL;
}


cEntityModel* cModel::addEntity(cEntity *entity, char* descriptor)
{
  cEntityModel* entmod = new cEntityModel(entity, this, this->createContext(descriptor));
  mEntities.push_back(entmod);
  return entmod;
}


void cModel::transformSelf()
{
  // Transform all entitymodel-states.
  cEntityModel *entmod;
  list<cEntityModel*>::iterator it;
  for (it = mEntities.begin(); it != mEntities.end();)
  {
    entmod = *it; it++;
    // Remove NULL-entities.
    if (entmod->mEntity == NULL) {
      mEntities.remove(entmod);
      continue;
    };
    // Remove Fragged-entities.
    if (entmod->mEntity->getFragged()) {
      mEntities.remove(entmod);
      continue;
    };
    // Transform this entitymodel-state.
    glPushMatrix();
    transformSelf(entmod);
    glPopMatrix();
  };
}


void cModel::transformSelf(cEntityModel *entmod)
{
}


void cModel::drawSelf()
{
  // Draw all entitymodel-states.
  cEntityModel *entmod;
  list<cEntityModel*>::iterator it;
  for (it = mEntities.begin(); it != mEntities.end();)
  {
    entmod = *it; it++;
    // Remove NULL-entities.
    if (entmod->mEntity == NULL) {
      mEntities.remove(entmod);
      continue;
    };
    // Remove Fragged-entities.
    if (entmod->mEntity->getFragged()) {
      mEntities.remove(entmod);
      continue;
    };
    // Draw this entitymodel-state.
    glPushMatrix();
    drawSelf(entmod);
    glPopMatrix();
  };
}


void cModel::drawSelf(cEntityModel *entmod)
{
}


void cModel::collideLine(float* a, float* b, cModelCollision* &result)
{
  cEntityModel *entmod;
  list<cEntityModel*>::iterator it;
  for (it = mEntities.begin(); it != mEntities.end();)
  {
    entmod = *it; it++;
		collideLine(entmod, a, b, result);
	};
}


void cModel::collideLine(cEntityModel *entmod, float* a, float* b, cModelCollision* &result)
{
}


int cModel::handleMsg(int msg, long prma, long prmb, long prmc)
{
  if (msg == 0)
  {
    char *_cmd = (char *) prma;
    string cmd = _cmd;
    string ref;
  };
  return ((cObject *)this)->handleMsg(msg, prma, prmb, prmc);
}


//--------------------------------


cEntity::cEntity()
{
  mController   = NULL;
}


void cEntity::animateSelf(float spf)
{
  if (mController) mController->processSelf();
}


int cEntity::handleMsg(int msg, long prma, long prmb, long prmc)
{
  if (msg == 0)
  {
    char *_cmd = (char *) prma;
    string cmd = _cmd;
    string ref;

    ref = "getRotPtr";
    if (cmd == ref) {
      float **f = ((float **) prmb);
      *f = mRot.mV;
      return 0;
    };

    ref = "getPosPtr";
    if (cmd == ref) {
      float **f = ((float **) prmb);
      *f = mPos.mV;
      return 0;
    };
  };
  return ((cObject *)this)->handleMsg(msg, prma, prmb, prmc);
}



//--------------------------------

