#ifndef OBJECTS_H
#define OBJECTS_H

#include "math3d.h"
#include "envir.h"
#include "control.h"

#undef distance

#include <list>
#include <cstdio>
#include <string>

using namespace std;

//-----------------------------------

	
class cWorld;

class cObject;
  class cModel;
  class cEntity;

class cEntityModel; //UML: cModel <-1--cEntityModel--*-> cEntity
	
class cModelCollision;

class cObjectFunction;



//-----------------------------------

class cWorld {

public: // Enums and Constant values.

//  enum { MILITARY_A, MILITARY_B, CIVILISTIC, NATURAL, PROJECTILE, WEAPON };
  enum { KILO, MEGA, GIGA, TERA, PETA, MAX_VISGRP };
//  enum { CALL_ALPHA, CALL_BRAVO, CALL_CHARLIE, CALL_DELTA, CALL_ECHO, MAX_CALLGRP };

public: // Deprecated, need to change:

  cDome* mDome; // Going to be a separated cObject.
  float mGravity[3]; // Downward force vector (or 0,0,0 in space).
//  float mWind[3]; // Wind force vector.

protected: // Attributes:

  list<cObject*> mObjects; // Contains *every* object which isn't yet completly deleted.

  list<cObject*> mModels; // Contains objects which are drawn and serve as models for other objects.

  list<cObject*> mCorpses[2]; // Doublebuffer for object delete scheduling.
  list<cObject*>* mBufferedCorpses; // Points to current one of the two delete doublebuffers.

//  map<string, void *> mGamevars;
//  map<string, string> mGamevars;

  float mTime; // Seconds since simulation begin.

  DWORD mOldClock; // Previous frame's time (for spf calculation).
  DWORD mNewClock; // Current frame's time (for spf calculation).

  float mSPF; // seconds per frame (1.0f/fps).
  float mFPS; // frames per second (1.0f/spf).

public: // Constructor, Object-Management and Drawing:

  cWorld();

  void spawnObject(cObject *object); // Spawns object imediatly.
  void fragObject(cObject *object); // Schedules object Frag for after next frame.
  void bagFragged(); // Call after each frame: Deletes fragged objects of previous frame.

  void updateTimers(); // Call every frame to update animation timers.

public: // Get/Set:

  float getSPF() { return mSPF; };
  float getFPS() { return mFPS; };
  float getTime() { return mTime; };

  float getGravity(int coord) { return mGravity[coord]; };
//  float getWind(int coord) { return mWind[coord]; };

public: // World-Scanning, World-Sense for objects.

//	ObjectFunction mechanism.
  bool callObjects(cObjectFunction* objfun, int groupMask, bool matchWholeMask);
};


//-----------------------------------


class cObject {
protected:

  string mName;
  cWorld* mWorld;
  bool mFragged; // Is this object scheduled to be deleted after the next frame?
  int mGroupFlags; // Bitmask to associate logical call and collision groups used in callObjects().
  int mCallGroup; // Determines order in which cObjects are called from callObjects (future).

public:
  cObject();

  void copySelf(cObject *source);
  virtual cObject* newSelf() { return new cObject; };
  virtual char* getClass() { return "cObject"; };
	virtual int getClassId() {return 0; }; // Used to identify major class interfaces.

  virtual void onSpawn();
  virtual void onFrag();

  string& getName() { return mName; };
  void setName(char *name) { mName = name; };

  cWorld* getWorld() { return mWorld; };
  void setWorld(cWorld* world) { mWorld = world; };

  bool getFragged() { return mFragged; };
  void setFragged() { mFragged = true; };
  
  int getGroupFlags() { return mGroupFlags; };
  void setGroupFlags(int flags) { mGroupFlags = flags; };

  int getCallGroup() { return mCallGroup; };
  void setCallGroup(int group) { mCallGroup = group; };

  virtual int handleMsg(int msg, long prma, long prmb, long prmc);
};


//-----------------------------------


class cModel : public cObject {
protected:
  list<cEntityModel*> mEntities;

  virtual void* createContext(char* descriptor = NULL);
  virtual void transformSelf(cEntityModel *entmod);
  virtual void drawSelf(cEntityModel *entmod);
  virtual void collideLine(cEntityModel *entmod, float* a, float* b, cModelCollision* &result);

public:
  cModel();
  virtual char* getClass() { return "cModel"; };
  virtual cObject* newSelf() { return new cModel; };
	int getClassId() {return 1; };

  cEntityModel* addEntity(cEntity *entity, char* descriptor = NULL);

  void transformSelf();
  void drawSelf();

  int handleMsg(int msg, long prma, long prmb, long prmc);

  void collideLine(float* a, float* b, cModelCollision* &result);
};


//-----------------------------------


class cEntity : public cObject {
public:
  cController* mController; // Controlled by this controller.
  cVector mPos;
  cVector mRot;

protected:
  cEntityModel* mModel;

public:
  cEntity();
  virtual char* getClass() { return "cEntity"; };
  virtual cObject* newSelf() { return new cEntity; };
	int getClassId() {return 2; };

  cController* getController() { return mController; };
  void setController(cController* controller) { mController = controller; };

  virtual void animateSelf(float spf);

  int handleMsg(int msg, long prma, long prmb, long prmc);
};


//-----------------------------------


class cModelCollision {
public:
  cEntityModel* intsec_entmod;
  float intsec_distance;
  float intsec_point[3];
  float intsec_normal[3];
};


//-----------------------------------


class cEntityModel {
public:
  cModel* mModel;
  cEntity* mEntity;
  void* mContext;

protected:

public:
  cEntityModel(cEntity *entity, cModel *model, void* context = NULL)
  {
    mModel = model;
    mEntity = entity;
    mContext = context;
  };
};


//-----------------------------------

class cObjectFunction {
public:
	cObject* caller;
	list<cObject*> collection;
	int amount;
public:
	cObjectFunction() { caller = NULL; amount = -1; };
	virtual void applyFunction(cObject *object) { };
	virtual bool moreObjects() { return true; };
};


class cObjectFunction_findName : public cObjectFunction {
public:
	cObject* caller;
	char* name;
public:
	cObjectFunction_findName(cObject* Exclude, const char *Name, int Amount) { caller = Exclude; name = (char*) Name; amount = Amount; };
	void applyFunction(cObject *object) { if ((amount != 0) && (object != caller) && (strcmp(object->getName().c_str(), name) == 0)) { amount--; collection.push_back(object); } };
	bool moreObjects() { if (amount == 0) return false; return true; };
};


class cObjectFunction_findClass : public cObjectFunction {
public:
	cObject* caller;
	char* classname;
public:
	cObjectFunction_findClass(const char *Classname, int Amount) { caller = NULL; classname = (char*) Classname; amount = Amount; };
	void applyFunction(cObject *object) { if ((amount != 0) && (object != caller) && (strcmp(object->getClass(), classname) == 0)) { amount--; collection.push_back(object); } };
	bool moreObjects() { if (amount == 0) return false; return true; };
};


class cObjectFunction_lineCollision : public cObjectFunction {
	public:
	float pointA[3];
	float pointB[3];
	cModelCollision* result;
	public:
	cObjectFunction_lineCollision(float* PointA, float* PointB)
	{
		pointA[0] = PointA[0]; pointA[1] = PointA[1]; pointA[2] = PointA[2];
		pointB[0] = PointB[0]; pointB[1] = PointB[1]; pointB[2] = PointB[2];
	};
	void applyFunction(cObject *object)
	{ /* If object is a model then call collision function passing <result> by reference. */
		if (object->getClassId() == 1)  ((cModel*) object)->collideLine(pointA, pointB, result);
	};
};


class cObjectFunction_sphereCollision : public cObjectFunction {
	public:
	float point[3];
	float radius;
	cModelCollision* result;
	public:
	cObjectFunction_sphereCollision(float* Point, float Radius) { point[0] = Point[0]; point[1] = Point[1]; point[2] = Point[2]; radius = Radius; };
	void applyFunction(cObject *object) { /* If object is a model then call collision function passing <result> by reference. */ };
};


class cObjectFunction_rangeScan : public cObjectFunction {
	public:
	float point[3];
	float range;
	public:
	cObjectFunction_rangeScan(float *Point, float Range, int Amount) { caller = NULL; point[0] = Point[0]; point[1] = Point[1]; point[2] = Point[2]; range = Range; amount = Amount; };
	void applyFunction(cObject *object)
	{ /* If object is an entity then ... */
		if (object->getClassId() == 2)
		{
			cEntity *ent = (cEntity*) object;
//			if (object != caller)
			if ((object != caller) && (lineDistance(point, ent->mPos.mV) < range))
			{
				amount--; collection.push_back(object);
			};
		};
	};
	bool moreObjects() { if (amount == 0) return false; return true; };
};


class cObjectFunction_animateEntity: public cObjectFunction {
	public:
	float spf;
	public:
	cObjectFunction_animateEntity(float SPF) { spf = SPF; };
	void applyFunction(cObject *object)
	{ /* If object is a model then ... */
		if (object->getClassId() == 2) ((cEntity*) object)->animateSelf(spf);
	};
};


class cObjectFunction_transformModel : public cObjectFunction {
	public:
	public:
	void applyFunction(cObject *object)
	{ /* If object is a model then ... */
		if (object->getClassId() == 1) { ((cModel*) object)->transformSelf(); };
	};
};


class cObjectFunction_drawModel : public cObjectFunction {
	public:
	public:
	void applyFunction(cObject *object)
	{ /* If object is a model then ... */
		if (object->getClassId() == 1) ((cModel*) object)->drawSelf();
	};
};

//-----------------------------------

#include "models.h"

//-----------------------------------

#include "entities.h"

//-----------------------------------


#endif
